﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Net;
using System.Collections.Generic;
using System.IO;

using Sencha.Direct;
using Sencha.Direct.Client;
using Sencha.Direct.Api;

namespace Sencha.Mvc.Tests
{


    /// <summary>
    ///This is a test class for DirectClientTest and is intended
    ///to contain all DirectClientTest Unit Tests
    ///</summary>
    [TestClass()]
    public class DirectClientTest
    {
        public class ResultCheck
        {
            public string Value;
            public bool CallbackHit;
            public Exception Exception;

        }

        private TestContext testContextInstance;

        /// <summary>
        ///Gets or sets the test context which provides
        ///information about and functionality for the current test run.
        ///</summary>
        public TestContext TestContext
        {
            get
            {
                return testContextInstance;
            }
            set
            {
                testContextInstance = value;
            }
        }

        #region Additional test attributes
        // 
        //You can use the following additional attributes as you write your tests:
        //
        //Use ClassInitialize to run code before running the first test in the class
        //[ClassInitialize()]
        //public static void MyClassInitialize(TestContext testContext)
        //{
        //}
        //
        //Use ClassCleanup to run code after all tests in a class have run
        //[ClassCleanup()]
        //public static void MyClassCleanup()
        //{
        //}
        //
        //Use TestInitialize to run code before running each test
        //[TestInitialize()]
        //public void MyTestInitialize()
        //{
        //}
        //
        //Use TestCleanup to run code after each test has run
        //[TestCleanup()]
        //public void MyTestCleanup()
        //{
        //}
        //
        #endregion


        private RpcConnection_Accessor Create_Accessor()
        {
            var instance = new PrivateObject(new RpcConnectionImpl(), new PrivateType(typeof(RpcConnection)));
            var target = new RpcConnection_Accessor(instance);
            return target;
        }

        private string GetTempFileUri(int tid)
        {
            var content = string.Format("[{{ \"type\":\"rpc\", \"tid\": {0}, \"action\":\"action\", \"method\":\"controller\", \"result\":\"hello\" }}]", tid);
            var tmp = Path.GetTempFileName();
            using (var fs = File.Open(tmp, FileMode.Create, FileAccess.Write, FileShare.None))
            {
                var bytes = System.Text.Encoding.UTF8.GetBytes(content);
                fs.Write(bytes, 0, bytes.Length);
                fs.Flush();
            }

            return "file://" + tmp;
        }

        /// <summary>
        ///A test for AbortRequest
        ///</summary>
        [TestMethod()]
        [DeploymentItem("Sencha.Mvc.dll")]
        public void EndRequestTest()
        {
            var target = Create_Accessor();
            var tempFile = this.GetTempFileUri(target.m_lastTransactionId + 1);

            target.m_routerUri = new Uri(tempFile);
            target.Batched = true;

            var args = new Dictionary<string, object>();
            var expectedResult = "hello";

            var resultCheck = new ResultCheck();
            resultCheck.Value = string.Empty;
            resultCheck.CallbackHit = false;
            var callback = new AsyncCallback(o =>
            {
                resultCheck.CallbackHit = true;
                resultCheck.Value = target.EndRequest(o) as string;
            });

            var asyncResult = target.BeginRequest("controller", "action", args, typeof(string), callback, null);

            Assert.IsNotNull(asyncResult, "Request does not return AsyncResuls");

            asyncResult.AsyncWaitHandle.WaitOne();

            Assert.IsTrue(resultCheck.CallbackHit, "Callback does not invoked");
            Assert.IsNotNull(resultCheck.Value, "Result is not parsed");
            Assert.AreEqual(expectedResult, resultCheck.Value);
        }

        /// <summary>
        ///A test for AbortRequest
        ///</summary>
        [TestMethod()]
        [DeploymentItem("Sencha.Mvc.dll")]
        public void AbortRequestTest()
        {
            var target = Create_Accessor();
            var tempFile = this.GetTempFileUri(target.m_lastTransactionId + 1);

            target.m_routerUri = new Uri(tempFile);
            target.Batched = false;

            RpcConnection.InternalTestsReadResponseDelay = 10000; // 1s delay then "Timeout" exception

            var args = new Dictionary<string, object>();

            var resultCheck = new ResultCheck();
            resultCheck.CallbackHit = false;
            var callback = new AsyncCallback(o =>
            {
                resultCheck.CallbackHit = true;
                target.EndRequest(o);

                Assert.Fail("Callback finished");
            });

            var asyncResult = target.BeginRequest("controller", "action", args, typeof(string), callback, null);

            Assert.IsNotNull(asyncResult, "Request does not return AsyncResuls");

            asyncResult.AsyncWaitHandle.WaitOne(1000);

            Assert.IsFalse(resultCheck.CallbackHit, "Callback invoked!");

            var aborted = target.AbortRequest(asyncResult);

            asyncResult.AsyncWaitHandle.WaitOne();

            Assert.IsTrue(aborted, "Request is not aborted");
        }

        [TestMethod()]
        [DeploymentItem("Sencha.Mvc.dll")]
        public void AbortRequestBatchedTest()
        {
            var target = Create_Accessor();
            var tempFile = this.GetTempFileUri(target.m_lastTransactionId + 1);

            target.m_routerUri = new Uri(tempFile);
            target.Batched = true;
            target.m_dispatchTimer.Change(-1, -1);  // disable dispatch timer

            var args = new Dictionary<string, object>();

            var callback = new AsyncCallback(o =>
            {
                Assert.Fail("Callback invoked");
            });

            var asyncResult = target.BeginRequest("controller", "action", args, typeof(string), callback, null);

            target.m_dispatchTimer.Change(-1, -1); // disable dispatch timer

            Assert.IsNotNull(asyncResult, "Request does not return AsyncResuls");

            var aborted = target.AbortRequest(asyncResult);

            Assert.IsTrue(aborted, "Request is not aborted");
        }

        /// <summary>
        ///A test for BeginRequest
        ///</summary>
        [TestMethod()]
        [DeploymentItem("Sencha.Mvc.dll")]
        public void RequestSyncTest()
        {
            var target = Create_Accessor();
            var tempFile = this.GetTempFileUri(target.m_lastTransactionId + 1);

            target.m_routerUri = new Uri(tempFile);
            target.Batched = false;

            var args = new Dictionary<string, object>();
            var result = string.Empty;
            var expectedResult = "hello";

            result = target.Request("controller", "action", args, typeof(string)) as string;

            Assert.AreEqual(expectedResult, result);
        }

        /// <summary>
        ///A test for BeginRequest
        ///</summary>
        [TestMethod()]
        [DeploymentItem("Sencha.Mvc.dll")]
        public void BeginRequestAsyncWithTimeoutTest()
        {
            var target = Create_Accessor();
            var tempFile = this.GetTempFileUri(target.m_lastTransactionId + 1);

            target.m_routerUri = new Uri(tempFile);
            target.Batched = true;
            RpcConnection.InternalTestsResponseDelay = 1000; // 1s delay then "Timeout" exception

            var args = new Dictionary<string, object>();

            var resultCheck = new ResultCheck();
            resultCheck.Exception = (Exception)null;
            resultCheck.CallbackHit = false;
            var callback = new AsyncCallback(o =>
            {
                resultCheck.CallbackHit = true;
                try
                {
                    target.EndRequest(o);
                }
                catch (Exception e)
                {
                    resultCheck.Exception = e;
                }
            });
            var asyncResult = target.BeginRequest("controller", "action", args, typeof(string), callback, null);

            Assert.IsNotNull(asyncResult, "Request does not return AsyncResuls");

            asyncResult.AsyncWaitHandle.WaitOne();

            Assert.IsNotNull(resultCheck.Exception, "No timeout exception was throwed");
            Assert.IsTrue(resultCheck.CallbackHit, "Callback does not invoked");
        }
    }
}
